@@ -16,9 +16,12 @@
 #include <sys/fanotify.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+#include <sys/vfs.h>
 #include <unistd.h>
 
 
+#define FUNUSED __attribute__((unused))
+
 PyDoc_STRVAR(ext__doc__,
 "Wrapper for fanotify.\n"
 "\n"
@@ -35,13 +38,49 @@ PyDoc_STRVAR(ext__doc__,
 #define AGAIN (errno == EAGAIN || errno == EWOULDBLOCK)
 #endif
 
+#define O_FLAGS (O_RDONLY|O_LARGEFILE|O_CLOEXEC|O_NOATIME)
+#define PID_CACHE_SIZE 64
+
+enum RUN_ERR_CODE {
+    PY_FILLED_ERR = -100,
+    FANO_MISMATCH_VER,
+};
+
+typedef struct c_rule c_rule_t;
+typedef struct _fs_list fs_list_t;
+
+typedef struct {
+    pid_t pid;
+    char our;
+} pid_cache_t;
+
+typedef struct {
+    c_rule_t *rules;
+    fs_list_t *fs_list;
+    pid_cache_t *cache_idx;
+    pid_cache_t pid_cache[PID_CACHE_SIZE];
+    long main_pid;
+    int fano_fd;
+    int log_fd;
+    int sock_fd;
+} fano_ctx_t;
+
 typedef struct str_val {
     uint32_t len;
     char buf[PATH_MAX];
 } str_val_t;
 
+typedef struct {
+    char *buf;
+    size_t bufsz;
+    size_t size;
+} buffer_t;
+
+#define BUFFER_INIT { .bufsz = 512 }
+
+// ### FMonRule ###############################################################
 struct c_rule {
-    struct c_rule *next;
+    c_rule_t *next;
     ssize_t hash;
     unsigned long ev_types;
     size_t pids_cnt;
@@ -93,7 +132,7 @@ typedef struct {
 })
 
 static void
-rules_list_add(struct c_rule **rules, ext_FanoRule *rule)
+rules_list_add(c_rule_t **rules, ext_FanoRule *rule)
 {
     ssize_t hash = PyObject_Hash(rule->name.obj);
     while (*rules) {
@@ -101,7 +140,7 @@ rules_list_add(struct c_rule **rules, ext_FanoRule *rule)
             return;
         rules = &(*rules)->next;
     }
-    struct c_rule *new;
+    c_rule_t *new;
     if (!(new = PyMem_Malloc(sizeof(*new))))
         return;
 
@@ -134,11 +173,11 @@ rules_list_add(struct c_rule **rules, ext_FanoRule *rule)
 }
 
 static void
-rules_list_raw_del(struct c_rule **rules, long hash)
+rules_list_raw_del(c_rule_t **rules, long hash)
 {
     while (*rules) {
         if (hash == (*rules)->hash) {
-            struct c_rule *next = (*rules)->next;
+            c_rule_t *next = (*rules)->next;
             PyMem_Free((*rules)->pids);
             PyMem_Free(*rules);
             *rules = next;
@@ -149,17 +188,17 @@ rules_list_raw_del(struct c_rule **rules, long hash)
 }
 
 static void
-rules_list_del(struct c_rule **rules, ext_FanoRule *rule)
+rules_list_del(c_rule_t **rules, ext_FanoRule *rule)
 {
     rules_list_raw_del(rules, PyObject_Hash(rule->name.obj));
 }
 
 static void
-rules_list_clear(struct c_rule **rules)
+rules_list_clear(c_rule_t **rules)
 {
-    struct c_rule *item = *rules;
+    c_rule_t *item = *rules;
     while (item) {
-        struct c_rule *next = item->next;
+        c_rule_t *next = item->next;
         PyMem_Free(item->pids);
         PyMem_Free(item);
         item = next;
@@ -168,7 +207,7 @@ rules_list_clear(struct c_rule **rules)
 }
 
 static unsigned char
-rule_pids_check(struct c_rule *rule, pid_t pid)
+rule_pids_check(c_rule_t *rule, pid_t pid)
 {
     if (rule->pids) {
         for (size_t i = 0; i < rule->pids_cnt; ++i)
@@ -309,9 +348,113 @@ static PyTypeObject ext_FanoRuleType = {
     .tp_members = FanoRule_members,
     .tp_methods = FanoRule_methods,
 };
+// ### ! FMonRule #############################################################
+
+// ### FMonRule ###############################################################
+struct _fs_list {
+    fs_list_t *next;
+    char *path;
+    fsid_t fsid;
+    int fd;
+};
+
+FUNUSED
+static void
+fs_list_add(fs_list_t **list, const char *path)
+{
+    while (*list) {
+        if (!strcmp((*list)->path, path))
+            // already exist
+            return;
+        list = &(*list)->next;
+    }
+
+    struct statfs buf;
+    if (statfs(path, &buf))
+        return;
+
+    fs_list_t *new = PyMem_Malloc(sizeof(*new));
+    if (!new || !(new->path = PyMem_Malloc(strlen(path) + 1))) {
+        PyMem_Free(new);
+        return;
+    }
+
+    new->next = NULL;
+    new->fd = open(path, O_RDONLY | O_DIRECTORY);
+    strcpy(new->path, path);
+    memcpy(&new->fsid, &buf.f_fsid, sizeof(new->fsid));
+
+    *list = new;
+}
+
+FUNUSED
+static void
+fs_list_del(fs_list_t **list, const char *path)
+{
+    while (*list) {
+        if (!strcmp((*list)->path, path)) {
+            fs_list_t *next = (*list)->next;
+            close((*list)->fd);
+            PyMem_Free((*list)->path);
+            PyMem_Free((*list));
+            *list = next;
+            return;
+        }
+        list = &(*list)->next;
+    }
+}
+
+static void
+fs_list_clear(fs_list_t **list)
+{
+    fs_list_t *item = *list;
+    while (item) {
+        fs_list_t *next = item->next;
+        close(item->fd);
+        PyMem_Free(item->path);
+        PyMem_Free(item);
+        item = next;
+    }
+    *list = 0;
+}
+
+FUNUSED
+static fs_list_t *
+fs_list_get_fs(fs_list_t **list, fsid_t *fsid)
+{
+    while (*list) {
+        if (!memcmp(&(*list)->fsid, fsid, sizeof(*fsid)))
+            return *list;
+        list = &(*list)->next;
+    }
+    return NULL;
+}
+// ### ! FMonRule #############################################################
+
+PyDoc_STRVAR(create__doc__,
+"create() -> int\n"
+"\n"
+"Create fanotify context\n"
+"\n"
+"Raises:\n"
+"    OSError\n");
+
+static PyObject *
+pyfanotify_create(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    fano_ctx_t *ctx = calloc(1, sizeof(*ctx));
+    if (!ctx)
+        return PyErr_SetFromErrno(PyExc_OSError);
+
+    ctx->fano_fd = ctx->log_fd = ctx->sock_fd = -1;
+    ctx->cache_idx = ctx->pid_cache;
+    ctx->main_pid = getpid();
+
+    return PyLong_FromVoidPtr(ctx);
+}
 
 PyDoc_STRVAR(init__doc__,
-"init(flags: int, o_flags: int) -> int\n"
+"init(ctx: int, flags: int, o_flags: int) -> int\n"
 "\n"
 "Wrapper for fanotify_init\n"
 "\n"
@@ -327,22 +470,29 @@ PyDoc_STRVAR(init__doc__,
 static PyObject *
 pyfanotify_init(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"flags", "o_flags", NULL};
+    static char *kwlist[] = {"ctx", "flags", "o_flags", NULL};
+    long long ctx_ptr;
     unsigned int flags, o_flags = 0;
+    fano_ctx_t *ctx;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "II:init", kwlist,
-                                     &flags, &o_flags))
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "LII:init", kwlist,
+                                     &ctx_ptr, &flags, &o_flags))
         return NULL;
 
-    int fd = fanotify_init(flags, o_flags);
-    if (fd == -1)
+    if (!(ctx = (void *)ctx_ptr)) {
+        PyErr_SetString(PyExc_ValueError, "Invalid context");
+        return NULL;
+    }
+
+    ctx->fano_fd = fanotify_init(flags, o_flags);
+    if (ctx->fano_fd == -1)
         return PyErr_SetFromErrno(PyExc_OSError);
 
-    return PyLong_FromLong(fd);
+    return PyLong_FromLong(ctx->fano_fd);
 }
 
 PyDoc_STRVAR(mark__doc__,
-"mark(fd: int, flags: int, mask: int, dirfd: int, pathname: str = None) -> None\n"
+"mark(ctx: int, flags: int, mask: int, dirfd: int, pathname: str = None) -> None\n"
 "\n"
 "Wrapper for fanotify_mark\n"
 "\n"
@@ -355,24 +505,39 @@ PyDoc_STRVAR(mark__doc__,
 static PyObject *
 pyfanotify_mark(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"fd", "flags", "mask", "dirfd", "pathname", NULL};
-    int fanotify_fd = -1;
+    static char *kwlist[] = {"ctx", "flags", "mask", "dirfd", "pathname", NULL};
+    long long ctx_ptr;
     unsigned int flags = 0;
     uint64_t mask = 0;
     int dirfd = -1;
     const char *pathname = NULL;
+    fano_ctx_t *ctx;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "iIKi|z:mark", kwlist,
-                                     &fanotify_fd, &flags, &mask, &dirfd, &pathname))
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "LIKi|z:mark", kwlist,
+                                     &ctx_ptr, &flags, &mask, &dirfd, &pathname))
         return NULL;
 
-    int err = fanotify_mark(fanotify_fd, flags, mask, dirfd, pathname);
+    if (!(ctx = (void *)ctx_ptr)) {
+        PyErr_SetString(PyExc_ValueError, "Invalid context");
+        return NULL;
+    }
+
+    int err = fanotify_mark(ctx->fano_fd, flags, mask, dirfd, pathname);
     if (err == -1) {
         if (pathname)
             return PyErr_SetFromErrnoWithFilename(PyExc_OSError, pathname);
         else
             return PyErr_SetFromErrno(PyExc_OSError);
     }
+#ifdef FAN_REPORT_FID
+    if (!(flags & (FAN_MARK_IGNORED_MASK | FAN_MARK_IGNORED_SURV_MODIFY))
+            && flags & (FAN_MARK_FILESYSTEM | FAN_MARK_ONLYDIR)) {
+        if (flags & FAN_MARK_ADD)
+            fs_list_add(&ctx->fs_list, pathname);
+        else
+            fs_list_del(&ctx->fs_list, pathname);
+    }
+#endif // FAN_REPORT_FID
 
     Py_RETURN_NONE;
 }
@@ -384,13 +549,14 @@ do_write(int fd, const void *data, size_t len)
     do {
         ret = write(fd, data, len);
     } while (ret < 0 && errno == EINTR);
+    fsync(fd);
     return ret;
 }
 
 static void
-do_log(int fd, const char *fmt, ...)
+do_log(fano_ctx_t *ctx, const char *fmt, ...)
 {
-    if (fd < 0)
+    if (ctx->log_fd < 0)
         return;
 
     char msg[4096];
@@ -407,57 +573,7 @@ do_log(int fd, const char *fmt, ...)
     if (len > sizeof(msg) - hdr_len)
         len = sizeof(msg) - hdr_len;
     *(uint32_t *)msg = len;
-    do_write(fd, &msg, len + hdr_len);
-}
-
-static int
-send_data(int sk_fd, struct c_rule *rule, const struct fanotify_event_metadata *ev,
-          str_val_t *exe, str_val_t *cwd, str_val_t *path)
-{
-    struct {
-        int64_t pid;
-        uint64_t ev_types;
-    } data = {ev->pid, ev->mask};
-    struct iovec iov[] = {
-        {&data, sizeof(data)},
-        {exe, exe->len + sizeof(exe->len)},
-        {cwd, cwd->len + sizeof(exe->len)},
-        {path, path->len + sizeof(exe->len)},
-    };
-    struct sockaddr_un addr = {.sun_family = AF_UNIX};
-    addr.sun_path[0] = '\0';
-
-    memcpy(addr.sun_path + 1, rule->name.buf, rule->name.len);
-    struct msghdr msg = {
-        .msg_name = &addr,
-        .msg_namelen = rule->name.len + offsetof(struct sockaddr_un, sun_path) + 1,
-        .msg_flags = 0,
-        .msg_iov = iov,
-        .msg_iovlen = sizeof(iov) / sizeof(*iov),
-    };
-    struct __attribute__((packed)) {
-        struct cmsghdr cm;
-        int fd;
-    } cmsg;
-    if (rule->pass_fd && ev->fd >= 0) {
-        cmsg.cm.cmsg_len = sizeof(cmsg);
-        cmsg.cm.cmsg_level = SOL_SOCKET;
-        cmsg.cm.cmsg_type = SCM_RIGHTS;
-        cmsg.fd = ev->fd;
-        msg.msg_control = &cmsg;
-        msg.msg_controllen = sizeof(cmsg);
-    }
-    for (int i = 0; i < 3; ++i) {
-        if (sendmsg(sk_fd, &msg, 0) < 0) {
-            if (AGAIN)
-                usleep(250000);
-            else if (errno != EINTR)
-                return errno;
-            continue;
-        }
-        break;
-    }
-    return 0;
+    do_write(ctx->log_fd, &msg, len + hdr_len);
 }
 
 static ssize_t
@@ -472,83 +588,281 @@ get_proc_str(const char *fmt, int meta, char *buf, size_t buf_size)
     return path_len;
 }
 
-#define FIND_RULE_CHECK_MATCH(name, fmt, meta) ({           \
-        if (rule->name##_pattern.len) {                     \
-            if (!(name).buf[0] &&                           \
-                !((name).len = get_proc_str(fmt, ev->meta,  \
-                        (name).buf, sizeof((name).buf))))   \
-                goto rules_cycle_continue;                  \
-            if (fnmatch(rule->name##_pattern.buf,           \
-                        (name).buf, FNM_EXTMATCH))          \
-                goto rules_cycle_continue;                  \
-        }                                                   \
-})
+static char *
+_readfd(buffer_t *bb, int fd)
+{
+    bb->size = 0;
+    ssize_t n;
+    size_t tot = 0;
+    do {
+        if ((tot >= bb->bufsz || !bb->buf) && !(bb->buf = realloc(bb->buf, bb->bufsz <<= 1)))
+            return 0;
+        if ((n = read(fd, &bb->buf[tot], bb->bufsz - tot)) != -1)
+            tot += n;
+        else if (errno != EINTR)
+            return 0;
+    } while (n);
+    bb->size = tot;
+    return bb->buf;
+}
+
+static pid_cache_t *
+pid_cache_check(fano_ctx_t *ctx, buffer_t *bb, pid_t pid)
+{
+    pid_cache_t *c = ctx->cache_idx;
+    for (unsigned i = PID_CACHE_SIZE; i--;) {
+        if (c->pid == pid)
+            return c;
+        if (++c >= &ctx->pid_cache[PID_CACHE_SIZE])
+            c = ctx->pid_cache;
+    }
+
+    if (c-- == ctx->pid_cache)
+        c = &ctx->pid_cache[PID_CACHE_SIZE - 1];
+    *c = (pid_cache_t){.pid = pid};
+
+    long p = (long)pid;
+    for (;;) {
+        if (p == ctx->main_pid) {
+            c->our = 1;
+            break;
+        }
+
+        char path[64], *ptr;
+        snprintf(path, sizeof(path), "/proc/%ld/stat", p);
+        int fd = open(path, O_RDONLY | O_CLOEXEC | O_NOCTTY);
+        if (fd == -1)
+            break;
+
+        char *buf = _readfd(bb, fd);
+        close(fd);
+        if (!buf)
+            break;
+        buf[bb->size] = 0;
+
+        if (!(ptr = strrchr(buf, ')')))
+            break;
+        ptr += 4;
+
+        // fields: https://man7.org/linux/man-pages/man5/proc.5.html
+        // field #4 - PPid
+        if (!(p = (long)strtoull(ptr, 0, 10)))
+            break;
+    }
+
+    return (ctx->cache_idx = c);
+}
 
 static int
-handle_events(int fd, struct c_rule **rules, int sk_fd, int log_fd)
+handle_events(fano_ctx_t *ctx)
 {
-    const struct fanotify_event_metadata *ev;
-    struct fanotify_event_metadata buf[8192];
+    struct fanotify_event_metadata buf[256];
     ssize_t len;
-    str_val_t exe, cwd, path, evt;
+    buffer_t bb = BUFFER_INIT;
+    int ret = 0;
 
-    if ((len = read(fd, buf, sizeof(buf))) == -1)
-        return errno;
-
-    ev = buf;
-    while (FAN_EVENT_OK(ev, len)) {
-        exe.buf[0] = cwd.buf[0] = path.buf[0] = evt.buf[0] = '\0';
-        exe.len = cwd.len = path.len = evt.len = 0;
-        if (ev->vers != FANOTIFY_METADATA_VERSION)
-            return -101;
-#ifdef FANO_PERM
-        unsigned char pass_fd = 0;
-        unsigned int resp = FAN_ALLOW;
-#endif // FANO_PERM
-        int sk_res;
-        for (struct c_rule *rule = *rules; rule;) {
-            if (rule_pids_check(rule, ev->pid))
-                goto rules_cycle_continue;
-            if (rule->ev_types && !(rule->ev_types & ev->mask))
-                goto rules_cycle_continue;
-            FIND_RULE_CHECK_MATCH(exe, "/proc/%d/exe", pid);
-            FIND_RULE_CHECK_MATCH(cwd, "/proc/%d/cwd", pid);
-            FIND_RULE_CHECK_MATCH(path, "/proc/self/fd/%d", fd);
-
-            if ((sk_res = send_data(sk_fd, rule, ev, &exe, &cwd, &path))) {
-                do_log(log_fd, "Fanotify: send_data error for %s: %s",
-                       rule->name.buf, strerror(sk_res));
-                if (sk_res == ECONNREFUSED) {
-                    do_log(log_fd, "Fanotify: delete \"%s\"", rule->name.buf);
-                    struct c_rule *to_del = rule;
-                    rule = rule->next;
-                    rules_list_raw_del(rules, to_del->hash);
-                    continue;
-                }
-            }
-#ifdef FANO_PERM
-            else
-                pass_fd |= rule->pass_fd;
-#endif // FANO_PERM
-            rules_cycle_continue:
-            rule = rule->next;
-        }
-        if (ev->fd != FAN_NOFD) {
-#ifdef FANO_PERM
-            if ((ev->mask & (FAN_OPEN_PERM|FAN_ACCESS_PERM)) && !pass_fd) {
-                struct fanotify_response r = {ev->fd, resp};
-                do_write(fd, &r, sizeof(r));
-            }
-#endif // FANO_PERM
-            close(ev->fd);
-        }
-        ev = FAN_EVENT_NEXT(ev, len);
+    if ((len = read(ctx->fano_fd, buf, sizeof(buf))) == -1) {
+        ret = errno;
+        goto end;
     }
-    return 0;
+
+    for (struct fanotify_event_metadata *ev = buf;
+            FAN_EVENT_OK(ev, len);
+            close(ev->fd), ev = FAN_EVENT_NEXT(ev, len)){
+        if (ev->vers != FANOTIFY_METADATA_VERSION) {
+            ret = FANO_MISMATCH_VER;
+            goto end;
+        }
+
+        if (pid_cache_check(ctx, &bb, ev->pid)->our)
+            // skip ours
+            continue;
+
+        int fnames = 0;
+        str_val_t exe[1], cwd[1], path[2], evt;
+        exe[0].buf[0] = cwd[0].buf[0] = path[0].buf[0] = path[1].buf[0] = evt.buf[0] = '\0';
+        exe[0].len = cwd[0].len = path[0].len = path[1].len = evt.len = 0;
+        c_rule_t to_del;
+
+#ifdef FAN_REPORT_FID
+        if (ev->event_len != FAN_EVENT_METADATA_LEN) {
+            struct fanotify_event_info_fid *fid;
+            struct fanotify_event_info_header *finfo = (struct fanotify_event_info_header *)(ev + 1);
+//            struct fanotify_event_info_pidfd *pidfd;
+//            struct fanotify_event_info_error *ierror;
+            struct file_handle *file_handle = NULL;
+            const char *file_name[2] = {};
+            ssize_t info_len = ev->event_len - ev->metadata_len;
+            ssize_t rest = info_len;
+            int ffd = FAN_NOFD, dfd = FAN_NOFD, *_fd = NULL;
+
+            int cont = 0;
+            while (rest) {
+                int with_fname = 1;
+                switch (finfo->info_type) {
+                case FAN_EVENT_INFO_TYPE_FID:
+                    with_fname = 0;
+# ifdef FAN_REPORT_DIR_FID
+                case FAN_EVENT_INFO_TYPE_DFID:
+                    with_fname = 0;
+                case FAN_EVENT_INFO_TYPE_DFID_NAME:
+# endif // FAN_REPORT_DIR_FID
+# ifdef FAN_RENAME
+                case FAN_EVENT_INFO_TYPE_OLD_DFID_NAME:
+                case FAN_EVENT_INFO_TYPE_NEW_DFID_NAME:
+# endif // FAN_RENAME
+                    fid = (struct fanotify_event_info_fid *)finfo;
+                    file_handle = (struct file_handle *)fid->handle;
+                    if (with_fname)
+                        file_name[fnames++] = (char *)(file_handle->f_handle + file_handle->handle_bytes);
+                    break;
+//                case FAN_EVENT_INFO_TYPE_PIDFD:
+//                    pidfd = (struct fanotify_event_info_pidfd *)finfo;
+//                    break;
+//                case FAN_EVENT_INFO_TYPE_ERROR:
+//                    ierror = (struct fanotify_event_info_error *)finfo;
+//                    break;
+                default:
+//                    do_log(ctx, "Fanotify: invalid info_type: 0x%02X\n", finfo->info_type);
+                    close(ffd);
+                    close(dfd);
+                    goto fid_end;
+                }
+
+# ifdef FAN_RENAME
+#  define FAN_FLAGS (FAN_CREATE|FAN_DELETE|FAN_MOVE|FAN_RENAME)
+# else
+#  define FAN_FLAGS (FAN_CREATE|FAN_DELETE|FAN_MOVE)
+# endif
+
+                if (ev->mask & FAN_FLAGS)
+                    _fd = &dfd;
+                else if (finfo->info_type == FAN_EVENT_INFO_TYPE_FID)
+                    _fd = &ffd;
+                else
+                    _fd = &dfd;
+
+                if (file_handle) {
+                    int evt_fd;
+                    fs_list_t *fs = fs_list_get_fs(&ctx->fs_list, (void *)&fid->fsid);
+                    if (!fs || (((evt_fd = open_by_handle_at(fs->fd, file_handle, O_FLAGS)) == FAN_NOFD)
+                                && (errno == ESTALE)))
+                        goto info_next;
+                    close(*_fd);
+                    *_fd = evt_fd;
+                }
+            info_next:
+                rest -= finfo->len;
+                finfo = (struct fanotify_event_info_header *)((uint8_t *)finfo + finfo->len);
+            }
+
+            if (ffd != FAN_NOFD) {
+                ev->fd = ffd;
+                close(dfd);
+                dfd = -1;
+            } else if (dfd != FAN_NOFD) {
+                if (fnames && !(ev->mask & FAN_FLAGS)) {
+                    ev->fd = openat(dfd, file_name[0], O_FLAGS);
+                    if (ev->fd == FAN_NOFD)
+                        ev->fd = dfd;
+                    else {
+                        close(dfd);
+                        dfd = FAN_NOFD;
+                    }
+                } else
+                    ev->fd = dfd;
+                if (fnames && ev->fd == dfd) {
+                    for (int i = 0; i < fnames; ++i) {
+                        path[i].len = (dfd >= 0) ? get_proc_str("/proc/self/fd/%d", dfd, path[i].buf, sizeof(path[i].buf)) : 0;
+                        path[i].buf[path[i].len] = '/';
+                        path[i].len = stpncpy(path[i].buf + path[i].len + 1, file_name[i], sizeof(path[i].buf) - path[i].len - 1) - path[i].buf;
+                    }
+                }
+            } else
+                cont = 1;
+        fid_end:
+            if (cont)
+                continue;
+        }
+# undef FAN_FLAGS
+#endif // FAN_REPORT_FID
+
+        for (c_rule_t *rule = ctx->rules; rule; rule = rule->next) {
+
+# define RULE_MATCH(name, fmt, meta, i)     \
+        (rule->name##_pattern.len           \
+            && (!(((name)[i]).buf[0]        \
+                    || (((name)[i]).len = get_proc_str(fmt, ev->meta, ((name)[i]).buf, sizeof(((name)[i]).buf))))  \
+                || fnmatch(rule->name##_pattern.buf, ((name)[i]).buf, FNM_EXTMATCH)))
+
+            if (rule_pids_check(rule, ev->pid)
+                    || (rule->ev_types && !(rule->ev_types & ev->mask))
+                    || RULE_MATCH(exe, "/proc/%ld/exe", pid, 0)
+                    || RULE_MATCH(cwd, "/proc/%ld/cwd", pid, 0)
+                    || RULE_MATCH(path, "/proc/self/fd/%ld", fd, 0)
+                    || RULE_MATCH(path, "/proc/self/fd/%ld", fd, 1))
+                continue;
+# undef RULE_MATCH
+
+            // sending data
+            struct {
+                int64_t pid;
+                uint64_t ev_types;
+            } data = {ev->pid, ev->mask};
+            struct iovec iov[] = {
+                    {&data, sizeof(data)},
+                    {exe,  exe[0].len + sizeof(exe[0].len)},
+                    {cwd,  cwd[0].len + sizeof(exe[0].len)},
+                    {&path[0], path[0].len + sizeof(exe[0].len)},
+                    {&path[1], path[1].len + sizeof(exe[0].len)},
+            };
+            struct sockaddr_un addr = {.sun_family = AF_UNIX};
+            memcpy(addr.sun_path + 1, rule->name.buf, rule->name.len);
+            struct msghdr msg = {
+                    .msg_name = &addr,
+                    .msg_namelen = rule->name.len + offsetof(struct sockaddr_un, sun_path) + 1,
+                    .msg_flags = 0,
+                    .msg_iov = iov,
+                    .msg_iovlen = (sizeof(iov) / sizeof(*iov)) - (2 - (fnames?:1)),
+            };
+            struct __attribute__((packed)) {
+                struct cmsghdr cm;
+                int fd;
+            } cmsg;
+            if (rule->pass_fd && ev->fd != FAN_NOFD) {
+                cmsg.cm.cmsg_len = sizeof(cmsg),
+                cmsg.cm.cmsg_level = SOL_SOCKET,
+                cmsg.cm.cmsg_type = SCM_RIGHTS,
+                cmsg.fd = ev->fd,
+                msg.msg_control = &cmsg;
+                msg.msg_controllen = sizeof(cmsg);
+            }
+
+            for (int i = 0; i < 3; ++i) {
+                if (sendmsg(ctx->sock_fd, &msg, 0) < 0) {
+                    if ((AGAIN && (usleep(250000) <= 0)) || errno == EINTR)
+                        continue;
+
+                    int e = errno;
+                    do_log(ctx, "FileMonitor: send_data error for %s: %s", rule->name.buf, strerror(e));
+                    if (e == ECONNREFUSED) {
+                        do_log(ctx, "FileMonitor: delete \"%s\"", rule->name.buf);
+                        to_del.next = rule->next;
+                        rules_list_raw_del(&ctx->rules, rule->hash);
+                        rule = &to_del;
+                    }
+                }
+                break;
+            }
+        }
+    }
+end:
+    free(bb.buf);
+    return ret;
 }
 
 PyDoc_STRVAR(run__doc__,
-"run(fd: int, rcon: Connection[, log_fd: int, fn, fn_args, fn_timeout=0]) -> None\n"
+"run(ctx: int, rcon: Connection[, log_fd: int, fn, fn_args, fn_timeout=0]) -> None\n"
 "\n"
 "Main routine. If the event matches the rule, information about the event\n"
 "will be sent to the unix socket named \"\\0\" + rule.name\n"
@@ -563,7 +877,7 @@ PyDoc_STRVAR(run__doc__,
 "        otherwise empty string;\n"
 "\n"
 "Args:\n"
-"    fd (int): Fanotify file descriptor.\n"
+"    ctx (int): Fanotify context.\n"
 "    rcon (Connection): Connection for read commands.\n"
 "    log_fd (int): Optionally. Logger file descriptor.\n"
 "        Message format:\n"
@@ -580,26 +894,36 @@ PyDoc_STRVAR(run__doc__,
 static PyObject *
 pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"fd", "rcon", "log_fd",
-                             "fn", "fn_args", "fn_timeout", 0};
-    int fd, rfd = -1, log_fd = -1, err = 0, sk = -1;
+    static char *kwlist[] = {"ctx", "rcon",
+                             "log_fd", "fn", "fn_args", "fn_timeout", 0};
+    long long ctx_ptr;
+    int rfd = -1, err = 0, log_fd = -1;
     time_t fn_timeout = 0;
     pid_t ppid = getppid();
-    struct c_rule *rules = NULL;
+    fano_ctx_t *ctx;
     PyObject *rcon, *tmp = NULL, *fn = NULL, *fn_args = NULL;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "iO|iOOl:run", kwlist,
-                                     &fd, &rcon, &log_fd, &fn, &fn_args, &fn_timeout))
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "LO|iOOl:run", kwlist,
+                                     &ctx_ptr, &rcon,
+                                     &log_fd, &fn, &fn_args, &fn_timeout))
         return 0;
 
-    if (fd < 0) {
+    if (!(ctx = (void *)ctx_ptr)) {
+        PyErr_SetString(PyExc_ValueError, "Invalid context");
+        err = PY_FILLED_ERR;
+        fn = fn_args = 0;
+        goto end;
+    }
+    ctx->log_fd = log_fd;
+
+    if (ctx->fano_fd < 0) {
         err = errno = EBADF;
         fn = fn_args = 0;
         goto end;
     }
     if (fn_timeout < 0) {
         PyErr_SetString(PyExc_ValueError, "timeout must be non-negative");
-        err = -100;
+        err = PY_FILLED_ERR;
         fn = fn_args = 0;
         goto end;
     }
@@ -608,7 +932,7 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
             Py_INCREF(fn);
         } else {
             fn = fn_args = NULL;
-            err = -100;
+            err = PY_FILLED_ERR;
             PyErr_Format(PyExc_TypeError, "'%.200s' object is not callable",
                          fn->ob_type->tp_name);
             goto end;
@@ -622,7 +946,7 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
     if (!((tmp = PyObject_CallMethod(rcon, "fileno", NULL))
             && ((rfd = (int)PyLong_AsLong(tmp)) >= 0))) {
         Py_XDECREF(tmp);
-        err = -100;
+        err = PY_FILLED_ERR;
         goto end;
     }
     Py_XDECREF(tmp);
@@ -630,10 +954,10 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
 
     struct pollfd fds[] = {
         {rfd, POLLIN, 0},
-        {fd, POLLIN, 0},
+        {ctx->fano_fd, POLLIN, 0},
     };
 
-    if ((sk = socket(PF_UNIX, SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK, 0)) == -1) {
+    if ((ctx->sock_fd = socket(PF_UNIX, SOCK_DGRAM | SOCK_CLOEXEC | SOCK_NONBLOCK, 0)) == -1) {
         err = errno;
         goto end;
     }
@@ -642,7 +966,7 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
     PyThreadState *state = PyEval_SaveThread();
     while (ppid == getppid()) {
         do {
-            if (rules && fn) {
+            if (ctx->rules && fn) {
                 if (fn_timeout) {
                     time_t tmp_time = time(0);
                     if (tmp_time - fn_timer < fn_timeout)
@@ -666,7 +990,7 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
             continue;
 
         if (fds[0].revents & POLLIN) {
-            struct c_rule *old = rules;
+            c_rule_t *old = ctx->rules;
             PyEval_RestoreThread(state);
 
             int cmd;
@@ -674,7 +998,7 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
             if (!((tmp = PyObject_CallMethod(rcon, "recv", NULL))
                     && PyArg_ParseTuple(tmp, "i|O", &cmd, &val))) {
                 Py_XDECREF(tmp);
-                err = -100;
+                err = PY_FILLED_ERR;
                 goto end;
             }
             switch (cmd) {
@@ -683,21 +1007,21 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
                 err = 0;
                 goto end;
             case CMD_CONNECT:
-                rules_list_add(&rules, (void *)val);
+                rules_list_add(&ctx->rules, (void *)val);
                 break;
             case CMD_DISCONNECT:
-                rules_list_del(&rules, (void *)val);
+                rules_list_del(&ctx->rules, (void *)val);
             default:
                 break;
             }
             Py_XDECREF(tmp);
 
             state = PyEval_SaveThread();
-            if (!old && rules)
+            if (!old && ctx->rules)
                 fn_timer = 0;
-            else if (old && !rules) {   // flush
-                fanotify_mark(fd, FAN_MARK_FLUSH, 0, AT_FDCWD, 0);
-                fanotify_mark(fd, FAN_MARK_FLUSH|FAN_MARK_MOUNT, 0, AT_FDCWD, 0);
+            else if (old && !ctx->rules) {   // flush
+                fanotify_mark(ctx->fano_fd, FAN_MARK_FLUSH, 0, AT_FDCWD, 0);
+                fanotify_mark(ctx->fano_fd, FAN_MARK_FLUSH | FAN_MARK_MOUNT, 0, AT_FDCWD, 0);
             }
 
         } else if (fds[0].revents & POLLNVAL) {
@@ -706,7 +1030,7 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
         }
 
         if (fds[1].revents & POLLIN) {
-            if ((err = handle_events(fds[1].fd, &rules, sk, log_fd)))
+            if ((err = handle_events(ctx)))
                 break;
         } else if (fds[1].revents & POLLNVAL) {
             err = errno = EBADF;
@@ -716,21 +1040,25 @@ pyfanotify_run(PyObject *self, PyObject *args, PyObject *kwargs)
     PyEval_RestoreThread(state);
 
 end:
-    close(sk);
-    rules_list_clear(&rules);
+    if (dup2(1, ctx->fano_fd) == -1)
+        close(ctx->fano_fd);
+    close(ctx->sock_fd);
+    rules_list_clear(&ctx->rules);
+    fs_list_clear(&ctx->fs_list);
+    free(ctx);
     Py_XDECREF(fn);
     Py_XDECREF(fn_args);
     switch (err) {
-        case 0:
-            Py_RETURN_NONE;
-        case -101:
-            PyErr_SetString(PyExc_AssertionError, "Mismatch of fanotify metadata version.");
-            return 0;
-        default:
-            errno = err;
-            PyErr_SetFromErrno(PyExc_OSError);
-        case -100:
-            return 0;
+    case 0:
+        Py_RETURN_NONE;
+    case FANO_MISMATCH_VER:
+        PyErr_SetString(PyExc_AssertionError, "Mismatch of fanotify metadata version.");
+        return 0;
+    default:
+        errno = err;
+        PyErr_SetFromErrno(PyExc_OSError);
+    case PY_FILLED_ERR:
+        return 0;
     }
 }
 
@@ -746,19 +1074,20 @@ end:
 
 
 static PyMethodDef ext_methods[] = {
-    {"init", (PyCFunction)pyfanotify_init, METH_VARARGS | METH_KEYWORDS, init__doc__},
-    {"mark", (PyCFunction)pyfanotify_mark, METH_VARARGS | METH_KEYWORDS, mark__doc__},
-    {"run", (PyCFunction)pyfanotify_run, METH_VARARGS | METH_KEYWORDS, run__doc__},
+        {"create", (PyCFunction)pyfanotify_create, METH_NOARGS, create__doc__},
+        {"init", (PyCFunction)pyfanotify_init, METH_VARARGS | METH_KEYWORDS, init__doc__},
+        {"mark", (PyCFunction)pyfanotify_mark, METH_VARARGS | METH_KEYWORDS, mark__doc__},
+        {"run", (PyCFunction)pyfanotify_run, METH_VARARGS | METH_KEYWORDS, run__doc__},
 //        {"response", (PyCFunction)pyfanotify_response, METH_VARARGS | METH_KEYWORDS, response__doc__},
-    {NULL, NULL, 0, NULL}
+        {NULL, NULL, 0, NULL}
 };
 
 static struct PyModuleDef moduledef = {
-    PyModuleDef_HEAD_INIT,
-    .m_name = "ext",
-    .m_doc = ext__doc__,
-    .m_size = -1,
-    .m_methods = ext_methods,
+        PyModuleDef_HEAD_INIT,
+        .m_name = "ext",
+        .m_doc = ext__doc__,
+        .m_size = -1,
+        .m_methods = ext_methods,
 };
 
 PyMODINIT_FUNC
@@ -808,9 +1137,9 @@ PyInit_ext(void)
     PyModule_AddIntMacro(module, FANOTIFY_METADATA_VERSION);
     PyModule_AddIntMacro(module, FAN_ALLOW);
     PyModule_AddIntMacro(module, FAN_DENY);
-    PyModule_AddIntMacro(module, FAN_AUDIT);
     PyModule_AddIntMacro(module, FAN_NOFD);
 
+    PyModule_AddIntMacro(module, O_CLOEXEC);
     PyModule_AddIntMacro(module, AT_FDCWD);
     PyModule_AddIntMacro(module, CMD_STOP);
     PyModule_AddIntMacro(module, CMD_CONNECT);
@@ -818,8 +1147,10 @@ PyInit_ext(void)
 
 #ifdef FAN_ENABLE_AUDIT    // (Linux 4.15)
     PyModule_AddIntMacro(module, FAN_ENABLE_AUDIT);
+    PyModule_AddIntMacro(module, FAN_AUDIT);
 #else
     PyModule_AddIntConstant(module, "FAN_ENABLE_AUDIT", 0);
+    PyModule_AddIntConstant(module, "FAN_AUDIT", 0);
 #endif // FAN_ENABLE_AUDIT (Linux 4.15)
 
 #ifdef FAN_REPORT_TID    // (Linux 4.20)
@@ -872,5 +1203,11 @@ PyInit_ext(void)
     PyModule_AddIntConstant(module, "FAN_REPORT_DFID_NAME", 0);
 #endif // FAN_REPORT_DIR_FID (Linux 5.9)
 
+#ifdef FAN_RENAME   // (Linux 5.17)
+    PyModule_AddIntMacro(module, FAN_RENAME);
+#else
+    PyModule_AddIntConstant(module, "FAN_RENAME", 0);
+#endif
+
     return module;
 }
